'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

/**
 * Creates an XHR/AJAX request object and returns it.  Accounts
 * for browsers that do not support or implement the
 * XMLHttpRequest.withCredentials property.
 *
 * @return {XMLHttpRequest}
 */
function createRequestObject () {
  // create our XHR variable
  var xhr = new XMLHttpRequest();

  // if we need send CORS requests on browsers that don't
  // support CORS using XMLHttpRequest then we need to look
  // for an alternative
  if (
    ! 'withCredentials' in xhr &&
    typeof XDomainRequest !== 'undefined'
  ) {
    xhr = new XDomainRequest();
  }

  return xhr;
}

/**
 * Creates a promise chain for the given XHR and request objects.
 * Pipes request and response values through the appropriate
 * interceptors and passes the error to each error handler if
 * needed.
 *
 * @param  {XMLHttpRequest} xhr
 * @param  {Object} req
 * @param  {Array} onReq
 * @param  {Array} onRes
 * @param  {Array} onErr
 * @return {Promise}
 */
function sendRequest (xhr, req, onReq, onRes, onErr) {
  if ( onReq === void 0 ) onReq = [];
  if ( onRes === void 0 ) onRes = [];
  if ( onErr === void 0 ) onErr = [];

  return Promise.resolve(req)
  .then(validateRequest)
  .then(applyInterceptors.bind(null, onReq, xhr))
  .then(sendRequestObject.bind(null, xhr))
  .then(formatResponse)
  .then(applyInterceptors.bind(null, onRes, xhr))
  .catch(invokeErrorHandlers.bind(null, onErr, xhr));
}

/**
 * Returns a promise for the onload and onerror events
 * of the given XHR object.
 *
 * @param  {XMLHttpRequest} xhr
 * @param  {Object} req
 * @return {Promise}
 */
function sendRequestObject (xhr, req) {
  return new Promise(function (accept, reject) {
    // process the onload event
    xhr.onload = function () {
      accept(this);
    };
    // process the fatal error event
    xhr.onerror = function () {
      reject(new Error('Portals was unable to make the network request!'));
    };
    // open our xhr request object and configure it
    xhr.open(req.method, req.url, true);
    xhr.withCredentials = req.withCredentials;
    // add the headers to the request
    for (var k in req.headers) {
      xhr.setRequestHeader(k, req.headers[k]);
    }
    // send the request
    xhr.send(req.body);
  });
}

/**
 * Pipes a value through each interceptor function while always
 * supplying the XHR object as the second argument to each one.
 * Returns the result of the promise chain.
 *
 * @param  {Array} interceptors
 * @param  {XMLHttpRequest} xhr
 * @param  {*} val
 * @return {Promise}
 */
function applyInterceptors (interceptors, xhr, val) {
  return interceptors.reduce(function (current, next) {
    return current.then(function (val) { return next(val, xhr); });
  }, Promise.resolve(val));
}

/**
 * Loops through each of the given error handlers, passing each one the
 * error object and the XHR object before returning a rejected promise
 * with the same error.
 *
 * @param  {Array} handlers
 * @param  {XMLHttpRequest} xhr
 * @param  {Error} err
 * @return {Promise}
 */
function invokeErrorHandlers (handlers, xhr, err) {
  handlers.forEach(function (handler) { return handler(err, xhr); });
  return Promise.reject(err);
}

/**
 * Formats a resolved XHR object as a simple object literal
 * with the status code, basic headers, response type and body
 * contents.
 *
 * @param  {XMLHttpRequest} xhr
 * @return {Object}
 */
function formatResponse (xhr) {
  return {
    status: xhr.status,
    headers: {
      'Content-Type': xhr.getResponseHeader('Content-Type'),
      'Cache-Control': xhr.getResponseHeader('Cache-Control'),
      'Expires': xhr.getResponseHeader('Expires')
    },
    type: xhr.responseType,
    body: xhr.responseText
  };
}

/**
 * Validates request options and ensures URL and method are provided.
 *
 * @param  {Object} req
 * @return {Object}
 */
function validateRequest (req) {
  // ensure method is present and a string
  if (typeof req.method !== 'string') {
    throw new Error('Invalid request method provided');
  }

  // ensure url is present and a string
  if (typeof req.url !== 'string') {
    throw new Error('Invalid url provided');
  }

  // ensure that "headers" is an object
  if (typeof req.headers !== 'object' || Array.isArray(req.headers)) {
    throw new Error('Invalid headers object provided');
  }

  return req;
}

/**
 * Replaces {tokens} with values from the given object literal
 * of key value pairs.
 *
 * @param  {String} str
 * @param  {Object} obj
 * @return {String}
 */
function parseTokens (str, obj) {
  return str.replace(/{ *([a-zA-Z0-9\-\_]+) *}/g, function (m, m1) {
    return obj[m1] || '{' + m1 + '}';
  });
}

/**
 * Creates a query string out of an object of key value pairs.
 *
 * @param  {Object} query
 * @return {String}
 */
function queryString (query) {
  return Object.keys(query).map(function (key, i) {
    // capture the value so we can modify if needed
    var val = query[key];
    // if the value is an array, comma-separate it
    if (Array.isArray(val)) val = val.join(',');
    // return the resulting query string
    return ((i === 0 ? '?' : '&') + key + '=' + val);
  }).join('');
}


var _utils = Object.freeze({
  createRequestObject: createRequestObject,
  sendRequest: sendRequest,
  sendRequestObject: sendRequestObject,
  applyInterceptors: applyInterceptors,
  invokeErrorHandlers: invokeErrorHandlers,
  formatResponse: formatResponse,
  validateRequest: validateRequest,
  parseTokens: parseTokens,
  queryString: queryString
});

/**
 * Simple JSON regex check.
 * @type {Regex}
 */
var isJson = /json/i;

/**
 * Higher-order function that will apply the given globals to
 * each request that gets piped through.
 *
 * @param  {Object} globals
 * @return {Function}
 */
function mergeGlobals (globals) {
  globals = Object.assign({
    hostname: '',
    headers: {
      Accept: 'application/json'
    }
  }, globals);

  return function (req) {
    var headers = (
      typeof req.headers !== 'object' ?
      Object.assign({}, globals.headers, req.headers) :
      globals.headers
    );

    req = Object.assign({}, globals, req);
    req.headers = headers;

    return req;
  }
}

/**
 * Returns a funciton that will build up the URL for a request
 * using the given url, a hostname (if set) and any params or
 * query objects.
 *
 * @param  {Bool} enableQueries
 * @param  {Bool} enableParams
 * @return {Function}
 */
function buildRequestUrl (enableQueries, enableParams) {
  return function (req) {
    // if "http" is not present in url, add the hostname
    if (req.url.indexOf('http') !== 0) {
      req.url = (req.hostname + '/' + req.url).replace(/\/+/g, '/');
    }

    // if params are present, switch out params with the values
    if (enableParams && typeof req.params === 'object') {
      req.url = parseTokens(req.url, req.params);
    }

    // build and append a query string if necessary
    if (
      enableQueries &&
      typeof req.query === 'object' &&
      ! Array.isArray(req.query)
    ) {
      req.url += queryString(req.query);
    }

    return req;
  }
}

/**
 * Encodes the request body as a JSON string when the body
 * is an object and Content-Type is set to "application/json".
 *
 * @param  {Object} req
 * @return {Object}
 */
function encodeJsonRequest (req) {
  if (
    typeof req.body === 'object' &&
    isJson.test(req.headers['Content-Type'])
  ) {
    req.body = JSON.stringify(req.body);
  }

  return req;
}

/**
 * Parses a JSON response body if Content-Type on the response
 * is set to "application/json".
 *
 * @param  {Object} res
 * @return {Object}
 */
function parseJsonResponse (res) {
  if (isJson.test(res.headers['Content-Type'])) {
    res.body = JSON.parse(res.body);
  }

 return res;
}

/**
 * Returns a function that will compare a given response's
 * "status" code against the given list of "ok statuses".  If
 * the list does not contain the status code, then the function
 * will throw an error.
 *
 * @param  {Array} allowed
 * @return {Function}
 */
function checkStatus (allowed) {
  if ( allowed === void 0 ) allowed = [];

  return function (res) {
    if (allowed.indexOf(res.status) === -1) {
      throw new Error('The response contained an invalid status code of ' + res.status);
    }

    return res;
  }
}

/**
 * Attempts to determine the "Content-Type" header for the request
 * based on the body value of the request.
 *
 * @param  {Object} req
 * @return {Object}
 */
function assumeContentType (req) {
  // ensure that we have headers object for req
  if ( ! req.headers) req.headers = {};
  // skip this if the header is explicitly set
  if ( ! req.headers['Content-Type']) {
    var type;
    // if the body is an instance of FormData, set type to multipart
    if (req.body instanceof FormData) type = 'multipart/form-data';
    // if the body starts with < assume XML
    else if (
      typeof req.body === 'string' && req.body.indexOf('<') === 0
    ) type = 'application/xml';
    // if the body is an object literal, set type to json
    else if (typeof req.body === 'object') type = 'application/json';
    // re-assign the Content-Type header
    if (type) req.headers['Content-Type'] = type;
  }

  return req;
}


var _int = Object.freeze({
  mergeGlobals: mergeGlobals,
  buildRequestUrl: buildRequestUrl,
  encodeJsonRequest: encodeJsonRequest,
  parseJsonResponse: parseJsonResponse,
  checkStatus: checkStatus,
  assumeContentType: assumeContentType
});

var Portal$1 = function Portal () {
  this._requestInterceptors = [];
  this._responseInterceptors = [];
  this._errorHandlers = [];
};

/**
 * Add an interceptor for modifying a request before it's sent
 * out via AJAX/XHR.
 *
 * @param{Function} fn
 * @return {this}
 */
Portal$1.prototype.onRequest = function onRequest (fn) {
  // enforce that the given value is a function
  if (typeof fn !== 'function') {
    throw new Error('You must provide a valid function as an interceptor!');
  }
  // add the interceptor to the list of interceptors
  this._requestInterceptors.push(fn);

  return this;
};

/**
 * Add an interceptor for modifying a response before it's given
 * to the user's promise subscriber.
 *
 * @param{Function} fn
 * @return {this}
 */
Portal$1.prototype.onResponse = function onResponse (fn) {
  // enforce that the given value is a function
  if (typeof fn !== 'function') {
    throw new Error('You must provide a valid function as an interceptor!');
  }
  // add the interceptor to the list of interceptors
  this._responseInterceptors.push(fn);

  return this;
};

/**
 * Add an error handler that will be invoked whenever a Portal
 * request errors out.
 *
 * @param{Function} fn
 * @return {this}
 */
Portal$1.prototype.onError = function onError (fn) {
  // enforce that the given value is a function
  if (typeof fn !== 'function') {
    throw new Error('You must provide a valid function as an error handler!');
  }
  // add the handler to the list of handlers
  this._errorHandlers.push(fn);

  return this;
};

/**
 * Creates an XHR object and sends it out, returning a promise
 * that can be used to subscribe to resolution or rejection
 * states.
 *
 * @param{Object} req
 * @return {Promise}
 */
Portal$1.prototype.send = function send (req) {
    if ( req === void 0 ) req = {};

  // create our XHR object for the request
  var xhr = createRequestObject();
  // send the request and get the promise for the result
  var promise = sendRequest(xhr, req, this._requestInterceptors, this._responseInterceptors, this._errorHandlers);
  // attach the XHR object's abort() method to the promise
  promise.abort = xhr.abort.bind(xhr);
  // return the promise
  return promise;
};

/**
 * Helper methods for GET, POST, PUT and DELETE.
 */

Portal$1.prototype.get = function get (url, req) {
    if ( req === void 0 ) req = {};

  this.send(Object.assign({}, req, { method: 'GET', url: url }));
};

Portal$1.prototype.post = function post (url, body, req) {
    if ( req === void 0 ) req = {};

  this.send(Object.assign({}, req, { method: 'POST', url: url, body: body }));
};

Portal$1.prototype.put = function put (url, body, req) {
    if ( req === void 0 ) req = {};

  this.send(Object.assign({}, req, { method: 'PUT', url: url, body: body }));
};

Portal$1.prototype.delete = function delete$1 (url, req) {
    if ( req === void 0 ) req = {};

  this.send(Object.assign({}, req, { method: 'DELETE', url: url }));
};

var OK_CODES = [
  200, 201, 202, 203, 204, 205, 206, 207, 208, 226,
  300, 301, 302, 303, 304, 305, 306, 307, 308
];

var utils = _utils;

var interceptors = _int;

var Portal = Portal$1;

/**
 * Creates a Portal instance and configures it with some default
 * interceptors based on the given options.
 *
 * @param  {Object} opts
 * @return {Portal}
 */
function createPortal (opts) {
  if ( opts === void 0 ) opts = {};

  // set up our options with our defaults underneath
  opts = Object.assign({
    globals: null,
    json: true,
    queries: true,
    params: true,
    okStatuses: OK_CODES,
    resolveTypes: true
  }, opts);

  // create our portal instance
  var portal = new Portal$1();

  // add globals interceptor if any globals are set
  portal.onRequest(mergeGlobals(opts.globals));

  // attempt to resolve Content-Types when enabled
  if (opts.resolveTypes) {
    portal.onRequest(assumeContentType);
  }

  // builds a complete url for each request
  portal.onRequest(
    buildRequestUrl(opts.queries, opts.params)
  );

  // if enabled, check the status codes of responses
  if (Array.isArray(opts.okStatuses)) {
    portal.onResponse(checkStatus(opts.okStatuses));
  }

  // add JSON encoding/decoding when enabled
  if (opts.json) {
    portal.onRequest(encodeJsonRequest);
    portal.onResponse(parseJsonResponse);
  }

  // return our configured portal instance!
  return portal;
}

exports.utils = utils;
exports.interceptors = interceptors;
exports.Portal = Portal;
exports.createPortal = createPortal;